import logging
from dataclasses import dataclass
from datetime import datetime, time
from typing import Optional

from pydantic import BaseModel, Extra, Field
from pydantic.functional_validators import PlainValidator
from typing_extensions import Annotated, TypeAliasType

from nagra_panorama_api.utils import el2dict


def first(iterable, default=None):
    return next(iter(iterable), default)


DATETIME_FORMAT = "%Y/%m/%d %H:%M:%S"


def parse_datetime(d):
    try:
        if d is None:
            return d
        return datetime.strptime(d, DATETIME_FORMAT)
    except Exception as e:
        logging.debug(e)
        logging.debug(f"Failed to parse {d} as datetime")
    return d


TIME_FORMAT = "%H:%M:%S"


def parse_time(d):
    return datetime.strptime(d, TIME_FORMAT).time()


def parse_tdeq(d):
    if "null" in d:
        return None
    try:
        return parse_time(d)
    except Exception as e:
        logging.debug(e)
    return parse_datetime(d)


def parse_progress(progress):
    try:
        return float(progress)
    except Exception as e:
        logging.debug(f"{e} => Fallback to datetime parsing")

    # When finished, progress becomes the date of the end
    if parse_datetime(progress):
        return 100.0
    return None


# https://docs.pydantic.dev/latest/concepts/types/#custom-types
# JobProgress = TypeAliasType('JobProgress', PlainValidator(parse_progress))
Datetime = TypeAliasType(
    "JobProgress", Annotated[datetime, PlainValidator(parse_progress)]
)


def single_xpath(xml, xpath, parser=None, default=None):
    try:
        res = xml.xpath(xpath)
        res = first(res)
    except Exception:
        return default
    if not res:
        return default
    if parser:
        res = parser(res)
    return res


pd = parse_datetime
sx = single_xpath


def mksx(xml):
    def single_xpath(xpath, parser=None, default=None):
        return sx(xml, xpath, parser=parser, default=default)

    return single_xpath


@dataclass
class SoftwareVersion:
    version: str
    filename: str
    released_on: datetime
    downloaded: bool
    current: bool
    latest: bool
    uploaded: bool

    @staticmethod
    def from_xml(xml):
        if isinstance(xml, (list, tuple)):
            xml = first(xml)
        if xml is None:
            return None
        p = mksx(xml)
        return SoftwareVersion(
            p("./version/text()"),
            p("./filename/text()"),
            p("./released-on/text()", parser=pd),
            p("./downloaded/text()") != "no",
            p("./current/text()") != "no",
            p("./latest/text()") != "no",
            p("./uploaded/text()") != "no",
        )

    @property
    def base_minor_version(self):
        major, minor, _ = self.version.split(".")
        return f"{major}.{minor}.0"

    @property
    def base_major_version(self):
        major, _, _ = self.version.split(".")
        return f"{major}.0.0"


@dataclass
class Job:
    tenq: datetime
    tdeq: time
    id: str
    user: str
    type: str
    status: str
    queued: bool
    stoppable: bool
    result: str
    tfin: datetime
    description: str
    position_in_queue: int
    progress: float
    details: str
    warnings: str

    @staticmethod
    def from_xml(xml):
        if isinstance(xml, (list, tuple)):
            xml = first(xml)
        if xml is None:
            return None
        p = mksx(xml)
        return Job(
            p("./tenq/text()", parser=pd),
            p("./tdeq/text()", parser=parse_tdeq),
            p("./id/text()"),
            p("./user/text()"),
            p("./type/text()"),
            p("./status/text()"),
            p("./queued/text()") != "NO",
            p("./stoppable/text()") != "NO",
            p("./result/text()"),
            p("./tfin/text()", parser=pd),
            p("./description/text()"),
            p("./positionInQ/text()", parser=int),
            p("./progress/text()", parser=parse_progress),
            "\n".join(xml.xpath("./details/line/text()")),
            p("./warnings/text()"),
        )


class Device(BaseModel):
    serial: str
    connected: bool
    unsupported_version: bool
    wildfire_rt: bool
    deactivated: Optional[str] = None
    hostname: Optional[str] = None
    ip_address: Optional[str] = None
    ipv6_address: Optional[str] = None
    mac_addr: Optional[str] = None
    uptime: Optional[str] = None
    family: Optional[str] = None
    model: Optional[str] = None
    sw_version: Optional[str] = None
    app_version: Optional[str] = None
    av_version: Optional[str] = None
    device_dictionary_version: Optional[str] = ""
    wildfire_version: Optional[str] = None
    threat_version: Optional[str] = None
    url_db: Optional[str] = None
    url_filtering_version: Optional[str] = None
    logdb_version: Optional[str] = None
    vpnclient_package_version: Optional[str] = None
    global_protect_client_package_version: Optional[str] = None
    prev_app_version: Optional[str] = None
    prev_av_version: Optional[str] = None
    prev_threat_version: Optional[str] = None
    prev_wildfire_version: Optional[str] = None
    prev_device_dictionary_version: Optional[str] = ""
    # domain/: str
    # slot_count: str
    # type/: str
    # tag/: str
    # plugin_versions
    # ha_cluster
    vpn_disable_mode: bool
    operational_mode: str
    certificate_status: Optional[str] = None
    certificate_subject_name: Optional[str] = None
    certificate_expiry: Datetime
    connected_at: Datetime
    custom_certificate_usage: Optional[str] = None
    multi_vsys: bool
    # vsys
    last_masterkey_push_status: str
    last_masterkey_push_timestamp: Optional[str] = None
    express_mode: bool
    device_cert_present: Optional[str] = None
    device_cert_expiry_date: str

    class Config:
        populate_by_name = True
        extra = Extra.allow
        # extra = Extra.ignore

    @staticmethod
    def from_xml(xml):
        if isinstance(xml, (list, tuple)):
            xml = first(xml)
        if xml is None:
            return None
        res = {k.replace("-", "_"): v for k, v in el2dict(xml)["entry"].items()}
        return Device(**res)


class VPNFlow(BaseModel):
    name: str
    id: int
    gwid: int
    inner_if: str
    outer_if: str
    localip: str
    peerip: str
    state: str
    mon: str
    owner: str

    class Config:
        populate_by_name = True
        extra = Extra.allow
        # extra = Extra.ignore

    @staticmethod
    def from_xml(xml):
        if isinstance(xml, (list, tuple)):
            xml = first(xml)
        if xml is None:
            return None
        res = {k.replace("-", "_"): v for k, v in el2dict(xml)["entry"].items()}
        return VPNFlow(**res)



import logging
from itertools import chain, product
from multiprocessing.pool import ThreadPool as Pool
from typing import Optional

import requests

from . import types
from .constants import SUCCESS_CODE
from .utils import (
    clean_url_host,
    detach,
    diff_patch,
    el2dict,
    etree_fromstring,
    etree_tostring,
    extend_element,
    map_dicts,
    wait,
)


def get_tree(
    host, api_key, remove_blank_text=True, verify=False, timeout=None, logger=None
):
    if logger is None:
        logger = logging
    res = requests.get(
        f"{host}/api?type=config&action=show&xpath=/config",
        headers={"X-PAN-KEY": api_key},
        verify=verify,
        timeout=timeout,
    )
    root_tree = etree_fromstring(res.content, remove_blank_text=remove_blank_text)
    try:
        tree = root_tree.xpath("/response/result/config")[0]
    except Exception:
        logger.warning("Response doesn't contains the config tag. Response:")
        logger.warning(etree_tostring(root_tree, pretty_print=True).decode()[:1000])
        raise Exception("Response doesn't contains the config tag. Check the logs")
    detach(tree)  # Remove the /response/result part
    return tree


def parse_msg_result(result):
    # sometimes, there is <line> elements inside of msg
    msg = "".join(result.xpath("./msg/text()"))
    if not msg:
        msg = "".join(result.xpath("./result/msg/text()"))
    if msg:
        return msg
    return "\n".join(result.xpath("./msg/line/text()"))


def _get_rule_use_cmd(device_group, position, rule_type, start_index, number):
    # positions = ("pre", "post")
    return f"""<show><policy-app>
        <mode>get-all</mode>
        <filter>(rule-state eq 'any')</filter>
        <vsysName>{device_group}</vsysName>
        <position>{position}</position>
        <type>{rule_type}</type>
        <anchor>{start_index}</anchor>
        <nrec>{number}</nrec>
        <pageContext>rule_usage</pageContext>
    </policy-app></show>"""


def raw_request(
    url,
    type,
    method="GET",
    vsys=None,
    params=None,
    headers=None,
    remove_blank_text=True,
    verify=False,
    logger=None,
):
    if logger is None:
        logger = logging
    query_params = {"type": type}
    if params:
        query_params = {**query_params, **params}
    if vsys is not None:
        query_params["vsys"] = vsys

    res = requests.request(
        method=method,
        url=url,
        params=query_params,
        headers=headers,
        verify=verify,
    )
    content = res.content.decode()
    tree = etree_fromstring(content, remove_blank_text=remove_blank_text)
    status = tree.attrib["status"]
    code = int(tree.get("code", SUCCESS_CODE))
    msg = parse_msg_result(tree)
    if status == "error" or code < SUCCESS_CODE:
        logger.debug(content[:500])
        raise Exception(msg)
    if msg:
        return msg
    return tree


class XMLApi:
    def __init__(self, host, api_key, verify=False, logger=None):
        if not host:
            raise Exception("Missing Host")
        host, _, _ = clean_url_host(host)

        self._host = host
        self._api_key = api_key
        self._url = f"{host}/api"
        self._verify = verify
        self.logger = logger or logging

    def _request(
        self,
        type,
        method="GET",
        vsys=None,
        params=None,
        remove_blank_text=True,
        verify=None,
    ):
        if verify is None:
            verify = self._verify
        headers = {"X-PAN-KEY": self._api_key}
        return raw_request(
            self._url,
            type,
            method,
            vsys=vsys,
            params=params,
            headers=headers,
            remove_blank_text=remove_blank_text,
            verify=verify,
            logger=self.logger,
        )

    def _conf_request(
        self,
        xpath,
        action="get",
        method="GET",
        vsys=None,
        params=None,
        remove_blank_text=True,
        verify=None,
    ):
        if params is None:
            params = {}
        params = {"action": action, "xpath": xpath, **params}
        return self._request(
            "config",
            method=method,
            vsys=vsys,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    def _op_request(
        self,
        cmd,
        method="POST",
        vsys=None,
        params=None,
        remove_blank_text=True,
        verify=None,
    ):
        if params is None:
            params = {}
        params = {"cmd": cmd, **params}
        return self._request(
            "op",
            method=method,
            vsys=vsys,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    def _commit_request(
        self,
        cmd,
        method="POST",
        params=None,
        remove_blank_text=True,
        verify=None,
    ):
        if params is None:
            params = {}
        params = {"cmd": cmd, **params}
        return self._request(
            "commit",
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/get-started-with-the-pan-os-xml-api/get-your-api-key
    def generate_apikey(self, username, password, verify=None):
        params = {"user": username, "password": password}
        return self._request(
            "keygen",
            method="POST",
            params=params,
            verify=verify,
        ).xpath(".//key/text()")[0]

    # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/pan-os-xml-api-request-types/get-version-info-api
    def api_version(self, verify=None):
        return el2dict(
            self._request(
                "version",
                method="POST",
                verify=verify,
            ).xpath(".//result")[0]
        )["result"]

    def configuration(
        self,
        xpath,
        action="get",
        method="GET",
        params=None,
        remove_blank_text=True,
        verify=None,
    ):
        return self._conf_request(
            xpath,
            action=action,
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    def operation(
        self,
        cmd,
        method="POST",
        params=None,
        remove_blank_text=True,
        verify=None,
    ):
        return self._op_request(
            cmd,
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    def get_tree(self, extended=False, verify=None):
        if verify is None:
            verify = self._verify
        tree = get_tree(self._host, self._api_key, verify=verify, logger=self.logger)
        if extended:
            self._extend_tree_information(tree, verify=verify)
        return tree

    def _get_rule_use(self, device_group, position, rule_type, number=200, verify=None):
        results = []
        for i in range(100):
            cmd = _get_rule_use_cmd(
                device_group,
                position,
                rule_type,
                i * number,
                number,
            )
            res = self._op_request(cmd, verify=verify).xpath("result")[0]
            total_count = int(res.attrib["total-count"])
            results.extend(res.xpath("entry"))
            if len(results) >= total_count:
                break
        return results

    def get_rule_use(self, tree=None, max_threads=None, verify=None):
        if tree is None:
            tree = self.get_tree(verify=verify)
        device_groups = tree.xpath("devices/*/device-group/*/@name")
        positions = ("pre", "post")
        # rule_types = tuple({x.tag for x in tree.xpath(
        # "devices/*/device-group/*"
        # "/*[self::post-rulebase or self::pre-rulebase]/*")})
        rule_types = ("security", "pbf", "nat", "application-override")
        args_list = list(product(device_groups, positions, rule_types))

        def func(args):
            return self._get_rule_use(*args, verify=verify)

        threads = len(args_list)
        threads = min(max_threads or threads, threads)
        with Pool(len(args_list)) as pool:
            data = pool.map(func, args_list)
        return [entry for entry_list in data for entry in entry_list]

    def _get_rule_hit_count(self, device_group, rulebase, rule_type, verify=None):
        cmd = (
            "<show><rule-hit-count><device-group>"
            f"<entry name='{device_group}'><{rulebase}><entry name='{rule_type}'>"
            f"<rules><all/></rules></entry></{rulebase}></entry>"
            "</device-group></rule-hit-count></show>"
        )
        res = self._op_request(cmd, verify=verify)
        entries = res.xpath(".//rules/entry") or []
        # return entries
        return [(device_group, rulebase, rule_type, e) for e in entries]

    def get_rule_hit_count(self, tree=None, max_threads=None, verify=None):
        if tree is None:
            tree = self.get_tree(verify=verify)
        device_groups = tree.xpath("devices/*/device-group/*/@name")
        # rulebases = tuple({x.tag for x in tree.xpath(
        # "devices/*/device-group/*/*")})
        rulebases = ("pre-rulebase", "post-rulebase")
        # rule_types = tuple({x.tag for x in tree.xpath(
        # "devices/*/device-group/*"
        # "/*[self::post-rulebase or self::pre-rulebase]/*")})
        rule_types = ("security", "pbf", "nat", "application-override")
        args_list = list(product(device_groups, rulebases, rule_types))

        def func(args):
            return self._get_rule_hit_count(*args, verify=verify)

        threads = len(args_list)
        threads = min(max_threads or threads, threads)
        with Pool(len(args_list)) as pool:
            data = pool.map(func, args_list)
        return [entry for entry_list in data for entry in entry_list]

    def _extend_tree_information(
        self,
        tree,
        extended=None,
        max_threads=None,
        verify=None,
    ):
        if extended is None:
            extended = self.get_rule_use(tree, max_threads=max_threads, verify=verify)
        rules = tree.xpath(
            ".//device-group/entry/"
            "*[self::pre-rulebase or self::post-rulebase]/*/rules/entry[@uuid]",
        )
        ext_dict = {x.attrib.get("uuid"): x for x in extended}
        rules_dict = {x.attrib["uuid"]: x for x in rules}
        for ext, rule in map_dicts(ext_dict, rules_dict):
            extend_element(rule, ext)
            # rule.extend(ext) # This is causing duplicates entries
        return tree, extended

    def get(self, xpath, verify=None):
        """
        This will retrieve the xml definition based on the xpath
        The xpath doesn't need to be exact
        and can select multiple values at once.
        Still, it must at least speciy /config at is begining
        """
        return self._conf_request(xpath, action="show", method="GET", verify=verify)

    def delete(self, xpath, verify=None):
        """
        This will REMOVE the xml definition at the provided xpath.
        The xpath must be exact.
        """
        return self._conf_request(
            xpath,
            action="delete",
            method="DELETE",
            verify=verify,
        )

    def create(self, xpath, xml_definition, verify=None):
        """
        This will ADD the xml definition
        INSIDE the element at the provided xpath.
        The xpath must be exact.
        """
        # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/pan-os-xml-api-request-types/configuration-api/set-configuration
        params = {"element": xml_definition}
        return self._conf_request(
            xpath,
            action="set",
            method="POST",
            params=params,
            verify=verify,
        )

    def update(self, xpath, xml_definition, verify=None):
        """
        This will REPLACE the xml definition
        INSTEAD of the element at the provided xpath
        The xpath must be exact.
        Nb: We can pull the whole config, update it locally,
        and push the final result
        """
        # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/pan-os-xml-api-request-types/configuration-api/set-configuration
        params = {"element": xml_definition}
        return self._conf_request(
            xpath,
            action="edit",
            method="POST",
            params=params,
            verify=verify,
        )

    def revert_changes(self, skip_validated=False, verify=None):
        skip = "<skip-validate>yes</skip-validate>" if skip_validated else ""
        cmd = f"<revert><config>{skip}</config></revert>"
        return self._op_request(cmd, verify=verify)

    def validate_changes(self, verify=None):
        cmd = "<validate><full></full></validate>"
        return self._op_request(cmd, verify=verify)

    def raw_get_push_scope(self, admin=None, verify=None):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        filter = f"<admin><member>{admin}</member></admin>" if admin else ""
        cmd = f"<show><config><push-scope>{filter}</push-scope></config></show>"
        return self._op_request(cmd, verify=verify)

    def get_push_scope_devicegroups(self, admin=None, verify=None):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        # TODO
        scope = self.raw_get_push_scope(admin=admin, verify=verify)
        return list(set(scope.xpath(".//objects/entry[@loc-type='device-group']/@loc")))

    def uncommited_changes(self, verify=None):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        cmd = "<show><config><list><changes></changes></list></config></show>"
        return self._op_request(cmd, verify=verify)

    def uncommited_changes_patch(self, verify=None):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        candidate = self.candidate_config(verify=verify)
        running = self.running_config(verify=verify)
        return diff_patch(running, candidate)

    def uncommited_changes_summary(self, admin=None, verify=None):
        """
        Only gives the concern device groups
        """
        admin = (
            f"<partial><admin><member>{admin}</member></admin></partial>"
            if admin
            else ""
        )
        cmd = f"<show><config><list><change-summary>{admin}</change-summary></list></config></show>"
        return self._op_request(cmd, verify=verify)

    def pending_changes(self, verify=None):
        """
        Result content is either 'yes' or 'no'
        """
        cmd = "<check><pending-changes></pending-changes></check>"
        return self._op_request(cmd, verify=verify)

    def candidate_config(self, verify=None):
        """
        Get the confirmation with pending changes
        """
        cmd = "<show><config><candidate></candidate></config></show>"
        return self._op_request(cmd, remove_blank_text=False, verify=verify)

    def running_config(self, verify=None):
        """
        Get the running configuration
        """
        cmd = "<show><config><running></running></config></show>"
        return self._op_request(cmd, remove_blank_text=False, verify=verify)

    def raw_get_jobs(self, job_id=None, verify=None):
        cmd = "<show><jobs>{}</jobs></show>".format(
            f"<id>{job_id}</id>" if job_id else "<all></all>",
        )
        return self._op_request(cmd, verify=verify)

    def raw_get_versions(self, verify=None):
        cmd = "<request><system><software><check></check></software></system></request>"
        return self.operation(cmd, verify=verify)

    def get_versions(self):
        res = self.raw_get_versions()
        return [
            types.SoftwareVersion.from_xml(entry)
            for entry in res.xpath(".//sw-updates/versions/entry")
        ]

    def get_jobs(self, job_id=None, verify=None):
        job_xmls = self.raw_get_jobs(job_id, verify=verify).xpath(".//job")
        res = [types.Job.from_xml(x) for x in job_xmls]
        if job_id:  # We want one specific ID
            return res[0]
        return res

    def wait_job_completion(self, job_id, waiter=None):
        if not waiter:
            waiter = wait()
        for _ in waiter:
            job = self.get_jobs(job_id)
            if job.progress >= 100:
                return job
            self.logger.info(f"Job {job_id} progress: {job.progress}")
        raise Exception("Timeout while waiting for job completion")

    def get_pending_jobs(self, verify=None):
        cmd = "<show><jobs><pending></pending></jobs></show>"
        return self._op_request(cmd, verify=verify)

    def commit_changes(self, force=False, verify=None):
        # force = "<force></force>"  # We don't want to support force option
        cmd = "<commit>{}</commit>".format("<force></force>" if force else "")
        return self._commit_request(cmd, verify=verify)

    def _lock_cmd(self, cmd, vsys, no_exception=False, verify=None):
        try:
            result = "".join(self._op_request(cmd, vsys=vsys, verify=verify).itertext())
            self.logger.debug(result)
        except Exception as e:
            if no_exception:
                self.logger.error(e)
                return False
            raise
        return True

    # https://github.com/PaloAltoNetworks/pan-os-python/blob/a6b018e3864ff313fed36c3804394e2c92ca87b3/panos/base.py#L4459
    def add_config_lock(
        self, comment=None, vsys="shared", no_exception=False, verify=None
    ):
        comment = f"<comment>{comment}</comment>" if comment else ""
        cmd = f"<request><config-lock><add>{comment}</add></config-lock></request>"
        return self._lock_cmd(cmd, vsys=vsys, no_exception=no_exception, verify=verify)

    def remove_config_lock(self, vsys="shared", no_exception=False, verify=None):
        cmd = "<request><config-lock><remove></remove></config-lock></request>"
        return self._lock_cmd(cmd, vsys=vsys, no_exception=no_exception, verify=verify)

    def add_commit_lock(
        self, comment=None, vsys="shared", no_exception=False, verify=None
    ):
        comment = f"<comment>{comment}</comment>" if comment else ""
        cmd = f"<request><commit-lock><add>{comment}</add></commit-lock></request>"
        return self._lock_cmd(cmd, vsys=vsys, no_exception=no_exception, verify=verify)

    def remove_commit_lock(self, vsys="shared", no_exception=False, verify=None):
        cmd = "<request><commit-lock><remove></remove></commit-lock></request>"
        return self._lock_cmd(cmd, vsys=vsys, no_exception=no_exception, verify=verify)

    def get_panorama_status(self, verify=None):
        cmd = "<show><panorama-status></panorama-status></show>"
        return self.operation(cmd, verify=verify).xpath(".//result")

    def raw_get_devices(self, connected=False, verify=None):
        # This only works on Panorama, not the FW
        filter = "<connected></connected>" if connected else "<all></all>"
        cmd = f"<show><devices>{filter}</devices></show>"
        return self.operation(cmd, verify=verify)

    def get_devices(self, connected=False, verify=None):
        entries = self.raw_get_devices(connected=connected, verify=verify).xpath(
            ".//devices/entry"
        )
        return [types.Device.from_xml(e) for e in entries]

    def raw_get_dg_hierarchy(self, verify=None):
        cmd = "<show><dg-hierarchy></dg-hierarchy></show>"
        return self.operation(cmd, verify=verify)

    def get_plan_dg_hierarchy(self, recursive=False, verify=None):
        devicegroups = {}  # name: children
        hierarchy = self.raw_get_dg_hierarchy(verify=verify).xpath(".//dg-hierarchy")[0]
        xpath = ".//dg" if recursive else "./dg"
        for dg in hierarchy.xpath(".//dg"):
            devicegroups[dg.attrib["name"]] = [
                x.attrib["name"] for x in dg.xpath(xpath)
            ]
        return devicegroups

    def raw_get_devicegroups(self, verify=None):
        cmd = "<show><devicegroups></devicegroups></show>"
        return self.operation(cmd, verify=verify)

    def get_devicegroups_name(
        self, parents=None, with_connected_devices=None, verify=None
    ):
        """
        This returns the names of the devicegroups:
        - parents: the returned list will only contain children of the provided parents (parents included)
        - with_devices: the returned list will only contain devicegroups that have direct devices under them
        """
        devicegroups = self.raw_get_devicegroups(verify=verify).xpath(
            ".//devicegroups/entry"
        )
        if with_connected_devices:
            names = [
                dg.attrib["name"]
                for dg in devicegroups
                if dg.xpath("./devices/entry/connected[text() = 'yes']")
            ]
        else:
            names = [dg.attrib["name"] for dg in devicegroups]
        if parents:
            hierarchy = self.get_plan_dg_hierarchy(recursive=True)
            tokeep = set(chain(*(hierarchy.get(p, []) for p in parents))) | set(parents)
            names = list(set(names) & tokeep)
        return names

    def raw_get_templates(self, name=None, verify=None):
        # This only works on Panorama, not the FW
        filter = f"<name>{name}</name>" if name else ""
        cmd = f"<show><templates>{filter}</templates></show>"
        return self.operation(cmd, verify=verify)

    def get_templates_sync_status(self, verify=None):
        res = self.raw_get_templates(verify=verify)
        for entry in res.xpath("./result/templates/entry"):
            template_name = entry.attrib["name"]
            for device in entry.xpath("./devices/entry"):
                device_name = device.attrib["name"]
                template_status = next(device.xpath("./template-status/text()"), None)
                yield (template_name, device_name, template_status)

    def raw_get_vpn_flows(self, name=None, verify=None):
        # This only works on Panorama, not the FW"
        filter = f"<name>{name}</name>" if name else "<all></all>"
        cmd = f"<show><vpn><flow>{filter}</flow></vpn></show>"
        return self.operation(cmd, verify=verify)

    def get_vpn_flows(self, name=None, verify=None):
        entries = self.raw_get_vpn_flows(name=name, verify=verify).xpath(
            ".//IPSec/entry"
        )
        return [types.VPNFlow.from_xml(e) for e in entries]

    def system_info(self, verify=None):
        cmd = "<show><system><info></info></system></show>"
        return self.operation(cmd, verify=verify)

    def raw_system_resources(self, verify=None):
        cmd = "<show><system><resources></resources></system></show>"
        return self.operation(cmd, verify=verify)

    def system_resources(self, verify=None):
        res = self.raw_system_resources(verify=verify)
        text = res.xpath(".//result/text()")[0]
        return text.split("\n\n")[0]

    def raw_download_software(self, version, verify=None):
        """
        version is the software version to download
        """
        cmd = f"<request><system><software><download><version>{version}</version></download></software></system></request>"
        return self.operation(cmd, verify=verify)

    def download_software(self, version, verify=None):
        """
        version is the software version to download
        """
        res = self.raw_download_software(version, verify=verify)
        try:
            return res.xpath(".//job/text()")[0]
        except Exception:
            self.logger.debug("Download has not started")
        return None

    def raw_install_software(self, version, verify=None):
        """
        version is the software version to install
        """
        cmd = f"<request><system><software><install><version>{version}</version></install></software></system></request>"
        return self.operation(cmd, verify=verify)

    def install_software(self, version, verify=None):
        """
        version is the software version to install
        """
        if isinstance(version, types.SoftwareVersion):
            version = version.version
        res = self.raw_install_software(version, verify=verify)
        try:
            return res.xpath(".//job/text()")[0]
        except Exception:
            self.logger.debug("Download has not started")
        return None

    def raw_restart(self, verify=None):
        cmd = "<request><restart><system></system></restart></request>"
        return self.operation(cmd, verify=verify)

    def restart(self, verify=None):
        return "".join(self.raw_restart(verify=verify).xpath(".//result/text()"))

    def automatic_download_software(self, version: Optional[str] = None):
        version_str = version
        versions = self.get_versions()
        if not version_str:
            version = next((v for v in versions if v.latest), None)
        else:
            version = next((v for v in versions if v.version == version_str), None)
        if not version:
            self.logger.error(f"Version {version_str} not found")
            return exit(1)

        # Already downloaded: Nothing to do
        if version.downloaded:
            self.logger.info(f"Version {version.version} already downloaded")
            return version

        # Download minor version first (required)
        base_version = next(
            (v for v in versions if v.version == version.base_minor_version), None
        )
        if not base_version.downloaded:
            self.logger.info(
                f"Launching download of minor version {base_version.version}"
            )
            job_id = self.download_software(base_version.version)
            if not job_id:
                raise Exception("Download has not started")
            job = self.wait_job_completion(job_id)
            if job.result != "OK":
                self.logger.debug(job)
                raise Exception(job.details)
            print(job.details)

        # Actually download the wanted version
        self.logger.info(f"Launching download of version {version.version}")
        job_id = self.download_software(version.version)
        if not job_id:
            raise Exception("Download has not started")
        job = self.wait_job_completion(job_id)
        if job.result != "OK":
            self.logger.debug(job)
            raise Exception(job.details)
        self.logger.info(job.details)
        return version

    def automatic_software_upgrade(
        self, version: Optional[str] = None, install=True, restart=True
    ):
        version = self.automatic_download_software(version)
        if version.current:
            self.logger.info(f"Version {version.version} is already installed")
            return version
        if not install:
            return version
        # We may get the following error:
        # "Error: Upgrading from 10.2.4-h10 to 11.1.2 requires a content version of 8761 or greater and found 8638-7689."
        # This should never happen, we decided to report the error and handle this manually
        self.logger.info(f"Launching install of version {version.version}")

        job_id = self.install_software(version.version)
        if not job_id:
            self.logger.error("Install has not started")
            exit(1)
        job = self.wait_job_completion(job_id)
        self.logger.info(job.details)

        if restart:
            self.logger.info("Restarting the device")
            restart_response = self.restart()
            self.logger.info(restart_response)
        return version



from panos import panorama

from .utils import clean_url_host


class Panorama(panorama.Panorama):
    def __init__(
        self,
        hostname,
        api_username=None,
        api_password=None,
        api_key=None,
        port=None,
        *args,
        **kwargs,
    ):
        _, hostname, _port = clean_url_host(hostname)
        port = port or _port or 443
        return super().__init__(
            hostname, api_username, api_password, api_key, port, *args, **kwargs
        )



"""
This example will download and install the latest version of the software.

Python requirements:
- ping3
"""

import logging
import os
from multiprocessing.pool import ThreadPool
from pathlib import Path

import click
import requests
import urllib3
import yaml

# from nagra_network_misc_utils.logger import set_default_logger
from rich.logging import RichHandler

# from ping3 import ping  # Ping
import nagra_panorama_api
from nagra_panorama_api.utils import wait

# set_default_logger()
logging.getLogger().setLevel(logging.INFO)
# logging.getLogger().setLevel(logging.DEBUG)


def wait_availability(client, fw_host, logger=None):
    if logger is None:
        logger = logging
    logger.info(
        f"Waiting for firewall '{fw_host}' availability. Repeating until success or timeout"
    )
    for _ in wait():
        try:
            versions = client.get_versions()
            current = next(v for v in versions if v.current)
            if current is None:
                logger.warning("Device is not not answering")
            else:
                return current
        except (urllib3.exceptions.MaxRetryError, requests.exceptions.ConnectionError):
            logger.warning(f"Firewall {fw_host} is still not responding")
        except Exception as e:
            logger.debug(
                f"Unexpected error of type {type(e)} occured on firewall {fw_host}"
            )
            logger.error(f"Firewall {fw_host} is still not responding: {e}")
    raise Exception(f"Timeout while waiting for availability of firewall {fw_host}")


def getenv(*varnames, exit_on_failure=True, default=None):
    for var in varnames:
        val = os.environ.get(var)
        if val is not None:
            return val.strip()
    logging.error(
        f"None of the following environment variables are defined: {', '.join(varnames)}"
    )
    if exit_on_failure:
        exit(1)
    return default


def check_availability(client, fw_host, logger=None):
    if logger is None:
        logger = logging
    ## Wait for the FW to respond
    version = wait_availability(client, fw_host, logger=logger)
    if not version:
        logger.error(f"Device {fw_host} never responded")
        return False
    logger.info(f"Firewall {fw_host} is available")
    return True


def check_panorama_connection(panorama_client, fw_host, logger=None):
    if logger is None:
        logger = logging
    ## Check that the device is connected to Panorama
    devices = panorama_client.get_devices()
    device = next((d for d in devices if d.ip_address == fw_host), None)
    if device is None:
        logger.error(f"Device with IP {fw_host} is not detected AT ALL to panorama")
        return False
    if not device.connected:
        logger.error(
            f"Device with IP {fw_host} is not detected as connected to panorama"
        )
        return False
    logger.info(f"Firewall {fw_host} is detected has connected to panorama")
    return True


def run_checks(panorama_client, client, fw_host, logger=None):
    if logger is None:
        logger = logging
    if not check_availability(client, fw_host, logger=logger):
        return False
    if not check_panorama_connection(panorama_client, fw_host, logger=logger):
        return False
    return True


################################################################################
DEFAULT_LOGGER_FORMAT = "%(message)s"


def get_logger(name, file, level=logging.INFO):
    logger = logging.getLogger(name)
    logger.handlers.clear()

    formatter = logging.Formatter(DEFAULT_LOGGER_FORMAT)
    stream_handler = RichHandler()
    logger.addHandler(stream_handler)
    if file is not None:
        file = Path(file).resolve()
        file.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(file, mode="w")
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    logger.setLevel(level)
    logger.propagate = False  # Prevent calling the root logger
    return logger


@click.command(
    "upgrade",
    help="Upgrade a group of devices to a different softare version",
)
@click.argument(
    "file",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    # help="Configuration file containing the upgrade groups"
)
@click.argument("group")  # help="Upgrade group from the configuration to run"
@click.option(
    "--apikey",
    help="Panorama & Firewall api key (must be the same for both). Use one of the following environment variables: PANORAMA_APIKEY, PANO_APIKEY",
)
@click.option(
    "--download-only",
    is_flag=True,
    default=False,
    help="Only download the software version without installing it",
)
@click.option("--test-only", is_flag=True, default=False, help="Only runs the checks")
@click.option("--logs-output", help="Destination folder where to put the logs")
def upgrade_cmd(file, group, apikey, download_only, test_only, logs_output):
    if download_only and test_only:
        logging.error(
            "You cannot not use --download-only and --test-only flags at the same time"
        )
        exit(1)
    file = Path(file).resolve()
    if not apikey:
        apikey = getenv("PANORAMA_APIKEY", "PANO_APIKEY")
    if file.suffix not in (".yml", ".yaml"):
        logging.error("File must be in yaml format")
        exit(1)
    if logs_output is not None:
        logs_output = Path(logs_output).resolve()
        if logs_output.is_file():
            logging.error(f"{logs_output} is a file, it must be a folder or not exists")
            exit(1)
        group_logs_output = logs_output / Path(group)
        if group_logs_output.is_file():
            logging.error(
                f"""\
Destination '{logs_output}' already contains a file named '{group}'.
Please rename/delete/move file '{group_logs_output}' or change the destination folder
"""
            )
            exit(1)
    try:
        data = yaml.safe_load(file.read_text())
        groups = data["groups"]
        group_data = groups.get(group)
        if not group_data:
            logging.error(f"Group '{group}' does not exist ini file {file}")
            defined_groups = groups.keys()
            if defined_groups:
                logging.info(
                    f"Possible groups are: {', '.join(sorted(defined_groups))}"
                )
            exit(1)
    except Exception as e:
        logging.error(
            f"""\
An error occured when parsing the file '{file}' and/or retrieving the data of group '{group}'
Make sure that the file is formatted correctly and that the group exists
{e}
"""
        )
        exit(1)

    panorama_ip = group_data["panorama_ip"]
    version = group_data.get("version")
    devices = group_data["devices"]

    if not test_only:
        if not version:
            logging.warning(
                "version not defined, the script will default to the latest version for each devices"
            )
        else:
            logging.info(f"Starting upgrade to version {version}")

    def worker(args):
        try:
            (
                group,
                panorama_host,
                fw_name,
                fw_host,
                apikey,
                version,
                download_only,
                test_only,
                logs_output,
            ) = args

            logfile = logs_output / Path(group) / fw_name if logs_output else None
            logger = get_logger(f"Device {fw_name}", logfile)

            panorama_client = nagra_panorama_api.XMLApi(
                panorama_host, apikey, logger=logger
            )
            client = nagra_panorama_api.XMLApi(fw_host, apikey, logger=logger)

            fw_sys_info = client.system_info()
            serial_number = fw_sys_info.xpath(".//serial/text()")[0]
            current_version = fw_sys_info.xpath(".//sw-version/text()")[0]
            logger.info(
                f"Device with IP {fw_host} (S/N: {serial_number}) is currently on version '{current_version}'"
            )
            if not test_only:
                version = client.automatic_software_upgrade(
                    version, install=(not download_only)
                )

            # Tests
            res = run_checks(panorama_client, client, fw_host, logger=logger)
            if download_only:
                logger.info(
                    f"Done dowloading version { version.version } on device {fw_name} ({fw_host}) "
                )
            elif not test_only:
                logger.info(
                    f"Done upgrading device {fw_name} ({fw_host}) to version { version.version }"
                )
            return res
        except Exception as e:
            logger.error(
                f"""\
An unexpected error occured
Worker args: {args}
{e}
"""
            )
            return False

    pool_data = [
        (
            group,
            panorama_ip,
            d["hostname"],
            d["ip_address"],
            apikey,
            version,
            download_only,
            test_only,
            logs_output,
        )
        for d in devices
    ]
    failed_upgrades = 0
    with ThreadPool(len(pool_data)) as pool:
        for i, upgrade_res in enumerate(pool.imap_unordered(worker, pool_data), 1):
            logging.info(f"Upraded {i}/{len(pool_data)}")
            if not upgrade_res:
                failed_upgrades += 1
    if failed_upgrades > 0:
        logging.error(f"There was {failed_upgrades} failed upgrades")
    else:
        logging.info("All upgrades went successfuly")


@click.command(
    "generate-apikey",
    help="This command generate the configuration file boilerplate using all devices connected to panorama",
)
@click.option(
    "--panorama-host",
    "host",
    envvar="PANORAMA_HOST",
    required=True,
    help="Panorama host/ip. Use PANORAMA_HOST environment variable",
)
@click.option(
    "--username",
    envvar="PANORAMA_USERNAME",
    required=True,
    help="Username password to access Panorama (envvar: PANORAMA_USERNAME)",
)
@click.option(
    "--password",
    envvar="PANORAMA_PASSWORD",
    prompt=True,
    hide_input=True,
    required=True,
    help="User password to access Panorama (envvar: PANORAMA_PASSWORD)",
)
def generate_apikey_cmd(host, username, password):
    client = nagra_panorama_api.XMLApi(host, "dummykey")
    print(client.generate_apikey(username, password))


@click.command(
    "generate-config",
    help="This command generate the configuration file boilerplate using all devices connected to panorama",
)
@click.option(
    "--panorama-host",
    "host",
    help="Panorama host/ip. Use PANORAMA_HOST environment variable",
)
@click.option(
    "--apikey",
    help="Panorama api key. Use one of the following environment variables: PANORAMA_APIKEY, PANO_APIKEY",
)
@click.option(
    "--sw-version",
    "version",
)
@click.option(
    "-o",
    "--out",
    help="Output file for the configuration containing the upgrade groups (default to stdin)",
)
def generate_configuration_file_cmd(host, apikey, version, out):
    if not host:
        host = getenv("PANORAMA_HOST")
    if not apikey:
        apikey = getenv("PANORAMA_APIKEY", "PANO_APIKEY")
    if not apikey:
        logging.error("Missing Panorama API Key")
        exit(1)
    panorama_client = nagra_panorama_api.XMLApi(host, apikey)
    connected_devices = panorama_client.get_devices(connected=True)
    data = {
        "groups": {
            "all": {
                "panorama_ip": host,
                "version": version,
                "devices": [
                    {
                        "hostname": d.hostname,
                        "ip_address": d.ip_address,
                    }
                    for d in connected_devices
                ],
            }
        },
    }

    content = yaml.dump(data, sort_keys=False, explicit_start=True, indent=2)
    if not out:
        print(content)
    else:
        output = Path(out)
        output.write_text(content)


# client = nagra_panorama_api.XMLApi(os.environ["PANO_HOST"], os.environ["PANO_APIKEY"])
# fw_sys_info = client.system_info()

# vpn_flows = client.get_vpn_flows()
# template_status = list(panorama_client.get_templates_sync_status())
